@@ -14,6 +14,7 @@
 
 import copy, os, re
 from collections import OrderedDict
+import itertools
 
 from . import environment
 from . import dependencies
@@ -28,23 +29,29 @@ known_basic_kwargs = {'install': True,
                       'c_pch': True,
                       'cpp_pch': True,
                       'c_args': True,
+                      'objc_args': True,
+                      'objcpp_args': True,
                       'cpp_args': True,
                       'cs_args': True,
                       'vala_args': True,
                       'fortran_args': True,
                       'd_args': True,
                       'java_args': True,
+                      'rust_args': True,
                       'link_args': True,
                       'link_depends': True,
                       'link_with': True,
                       'link_whole': True,
+                      'implicit_include_directories': True,
                       'include_directories': True,
                       'dependencies': True,
                       'install_dir': True,
                       'main_class': True,
+                      'name_suffix': True,
                       'gui_app': True,
                       'extra_files': True,
                       'install_rpath': True,
+                      'build_rpath': True,
                       'resources': True,
                       'sources': True,
                       'objects': True,
@@ -62,14 +69,17 @@ known_lib_kwargs = known_basic_kwargs.copy()
 known_lib_kwargs.update({'version': True, # Only for shared libs
                          'soversion': True, # Only for shared libs
                          'name_prefix': True,
-                         'name_suffix': True,
                          'vs_module_defs': True, # Only for shared libs
                          'vala_header': True,
                          'vala_vapi': True,
                          'vala_gir': True,
                          'pic': True, # Only for static libs
+                         'rust_crate_type': True, # Only for Rust libs
                          })
 
+known_exe_kwargs = known_basic_kwargs.copy()
+known_exe_kwargs.update({'implib': True,
+                         })
 
 class InvalidArguments(MesonException):
     pass
@@ -115,7 +125,7 @@ class Build:
             self.compilers[lang] = compiler
 
     def add_cross_compiler(self, compiler):
-        if len(self.cross_compilers) == 0:
+        if not self.cross_compilers:
             self.static_cross_linker = self.environment.detect_static_linker(compiler)
         lang = compiler.get_language()
         if lang not in self.cross_compilers:
@@ -226,8 +236,6 @@ class ExtractedObjects:
               'You can only extract all the object files at once.'
         raise MesonException(msg)
 
-    def get_want_all_objects(self):
-        return self.want_all_objects
 
 class EnvironmentVariables:
     def __init__(self):
@@ -237,7 +245,7 @@ class EnvironmentVariables:
         repr_str = "<{0}: {1}>"
         return repr_str.format(self.__class__.__name__, self.envvars)
 
-    def get_value(self, name, values, kwargs):
+    def get_value(self, values, kwargs):
         separator = kwargs.get('separator', os.pathsep)
 
         value = ''
@@ -246,16 +254,16 @@ class EnvironmentVariables:
         return separator, value.strip(separator)
 
     def set(self, env, name, values, kwargs):
-        return self.get_value(name, values, kwargs)[1]
+        return self.get_value(values, kwargs)[1]
 
     def append(self, env, name, values, kwargs):
-        sep, value = self.get_value(name, values, kwargs)
+        sep, value = self.get_value(values, kwargs)
         if name in env:
             return env[name] + sep + value
         return value
 
     def prepend(self, env, name, values, kwargs):
-        sep, value = self.get_value(name, values, kwargs)
+        sep, value = self.get_value(values, kwargs)
         if name in env:
             return value + sep + env[name]
 
@@ -269,6 +277,11 @@ class EnvironmentVariables:
 
 class Target:
     def __init__(self, name, subdir, build_by_default):
+        if '/' in name or '\\' in name:
+            # Fix failing test 53 when this becomes an error.
+            mlog.warning('''Target "%s" has a path separator in its name.
+This is not supported, it can cause unexpected failures and will become
+a hard error in the future.''' % name)
         self.name = name
         self.subdir = subdir
         self.build_by_default = build_by_default
@@ -340,9 +353,7 @@ class BuildTarget(Target):
         self.process_objectlist(objects)
         self.process_kwargs(kwargs, environment)
         self.check_unknown_kwargs(kwargs)
-        if len(self.sources) == 0 \
-                and len(self.generated) == 0 \
-                and len(self.objects) == 0:
+        if not self.sources and not self.generated and not self.objects:
             raise InvalidArguments('Build target %s has no sources.' % name)
         self.process_compilers()
         self.validate_sources()
@@ -433,7 +444,7 @@ class BuildTarget(Target):
         We also add compilers that were used by extracted objects to simplify
         dynamic linker determination.
         '''
-        if len(self.sources) + len(self.generated) + len(self.objects) == 0:
+        if not self.sources and not self.generated and not self.objects:
             return
         # Populate list of compilers
         if self.is_cross:
@@ -488,7 +499,7 @@ class BuildTarget(Target):
             self.compilers['c'] = compilers['c']
 
     def validate_sources(self):
-        if len(self.sources) == 0:
+        if not self.sources:
             return
         for lang in ('cs', 'java'):
             if lang in self.compilers:
@@ -505,6 +516,34 @@ class BuildTarget(Target):
                 assert(len(self.compilers) == 1)
                 return
 
+    def process_link_depends(self, sources, environment):
+        """Process the link_depends keyword argument.
+
+        This is designed to handle strings, Files, and the output of Custom
+        Targets. Notably it doesn't handle generator() returned objects, since
+        adding them as a link depends would inherently cause them to be
+        generated twice, since the output needs to be passed to the ld_args and
+        link_depends.
+        """
+        if not isinstance(sources, list):
+            sources = [sources]
+        for s in sources:
+            if hasattr(s, 'held_object'):
+                s = s.held_object
+
+            if isinstance(s, File):
+                self.link_depends.append(s)
+            elif isinstance(s, str):
+                self.link_depends.append(
+                    File.from_source_file(environment.source_dir, self.subdir, s))
+            elif hasattr(s, 'get_outputs'):
+                self.link_depends.extend(
+                    [File.from_built_file(s.subdir, p) for p in s.get_outputs()])
+            else:
+                raise InvalidArguments(
+                    'Link_depends arguments must be strings, Files, '
+                    'or a Custom Target, or lists thereof.')
+
     def get_original_kwargs(self):
         return self.kwargs
 
@@ -513,10 +552,14 @@ class BuildTarget(Target):
             d = [d]
         newd = []
         for i in d:
-            if hasattr(i, 'held_object'):
-                newd.append(i.held_object)
-            else:
-                newd.append(i)
+            if isinstance(i, list):
+                i = self.unpack_holder(i)
+            elif hasattr(i, 'held_object'):
+                i = i.held_object
+            for t in ['dependencies', 'link_with', 'include_directories', 'sources']:
+                if hasattr(i, t):
+                    setattr(i, t, self.unpack_holder(getattr(i, t)))
+            newd.append(i)
         return newd
 
     def copy_kwargs(self, kwargs):
@@ -524,10 +567,14 @@ class BuildTarget(Target):
         # This sucks quite badly. Arguments
         # are holders but they can't be pickled
         # so unpack those known.
-        if 'dependencies' in self.kwargs:
-            self.kwargs['dependencies'] = self.unpack_holder(self.kwargs['dependencies'])
-        if 'link_with' in self.kwargs:
-            self.kwargs['link_with'] = self.unpack_holder(self.kwargs['link_with'])
+        for k, v in self.kwargs.items():
+            if isinstance(v, list):
+                self.kwargs[k] = self.unpack_holder(v)
+            if hasattr(v, 'held_object'):
+                self.kwargs[k] = v.held_object
+        for t in ['dependencies', 'link_with', 'include_directories', 'sources']:
+            if t in self.kwargs:
+                self.kwargs[t] = self.unpack_holder(self.kwargs[t])
 
     def extract_objects(self, srclist):
         obj_src = []
@@ -602,28 +649,37 @@ class BuildTarget(Target):
         if not isinstance(valalist, list):
             valalist = [valalist]
         self.add_compiler_args('vala', valalist)
+        objclist = kwargs.get('objc_args', [])
+        if not isinstance(objclist, list):
+            objclist = [objclist]
+        self.add_compiler_args('objc', objclist)
+        objcpplist = kwargs.get('objcpp_args', [])
+        if not isinstance(objcpplist, list):
+            objcpplist = [objcpplist]
+        self.add_compiler_args('objcpp', objcpplist)
         fortranlist = kwargs.get('fortran_args', [])
         if not isinstance(fortranlist, list):
             fortranlist = [fortranlist]
         self.add_compiler_args('fortran', fortranlist)
+        rustlist = kwargs.get('rust_args', [])
+        if not isinstance(rustlist, list):
+            rustlist = [rustlist]
+        self.add_compiler_args('rust', rustlist)
         if not isinstance(self, Executable):
             self.vala_header = kwargs.get('vala_header', self.name + '.h')
             self.vala_vapi = kwargs.get('vala_vapi', self.name + '.vapi')
             self.vala_gir = kwargs.get('vala_gir', None)
         dlist = stringlistify(kwargs.get('d_args', []))
         self.add_compiler_args('d', dlist)
-        self.link_args = kwargs.get('link_args', [])
-        if not isinstance(self.link_args, list):
-            self.link_args = [self.link_args]
+        self.link_args = flatten(kwargs.get('link_args', []))
         for i in self.link_args:
             if not isinstance(i, str):
                 raise InvalidArguments('Link_args arguments must be strings.')
-        self.link_depends = kwargs.get('link_depends', [])
-        if not isinstance(self.link_depends, list):
-            self.link_depends = [self.link_depends]
-        for i in self.link_depends:
-            if not isinstance(i, str):
-                raise InvalidArguments('Link_depends arguments must be strings.')
+        for l in self.link_args:
+            if '-Wl,-rpath' in l or l.startswith('-rpath'):
+                mlog.warning('''Please do not define rpath with a linker argument, use install_rpath or build_rpath properties instead.
+This will become a hard error in a future Meson release.''')
+        self.process_link_depends(kwargs.get('link_depends', []), environment)
         # Target-specific include dirs must be added BEFORE include dirs from
         # internal deps (added inside self.add_deps()) to override them.
         inclist = kwargs.get('include_directories', [])
@@ -650,18 +706,20 @@ class BuildTarget(Target):
         elif 'gui_app' in kwargs:
             raise InvalidArguments('Argument gui_app can only be used on executables.')
         extra_files = kwargs.get('extra_files', [])
-        if isinstance(extra_files, str):
+        if not isinstance(extra_files, list):
             extra_files = [extra_files]
         for i in extra_files:
-            if not isinstance(i, str):
-                raise InvalidArguments('Arguments to extra_files must be strings.')
-            trial = os.path.join(environment.get_source_dir(), self.subdir, i)
+            assert(isinstance(i, File))
+            trial = os.path.join(environment.get_source_dir(), i.subdir, i.fname)
             if not(os.path.isfile(trial)):
                 raise InvalidArguments('Tried to add non-existing extra file %s.' % i)
         self.extra_files = extra_files
         self.install_rpath = kwargs.get('install_rpath', '')
         if not isinstance(self.install_rpath, str):
             raise InvalidArguments('Install_rpath is not a string.')
+        self.build_rpath = kwargs.get('build_rpath', '')
+        if not isinstance(self.build_rpath, str):
+            raise InvalidArguments('Build_rpath is not a string.')
         resources = kwargs.get('resources', [])
         if not isinstance(resources, list):
             resources = [resources]
@@ -675,7 +733,7 @@ class BuildTarget(Target):
         if 'name_prefix' in kwargs:
             name_prefix = kwargs['name_prefix']
             if isinstance(name_prefix, list):
-                if len(name_prefix) != 0:
+                if name_prefix:
                     raise InvalidArguments('name_prefix array must be empty to signify null.')
             elif not isinstance(name_prefix, str):
                 raise InvalidArguments('name_prefix must be a string.')
@@ -684,7 +742,7 @@ class BuildTarget(Target):
         if 'name_suffix' in kwargs:
             name_suffix = kwargs['name_suffix']
             if isinstance(name_suffix, list):
-                if len(name_suffix) != 0:
+                if name_suffix:
                     raise InvalidArguments('name_suffix array must be empty to signify null.')
             else:
                 if not isinstance(name_suffix, str):
@@ -704,6 +762,9 @@ class BuildTarget(Target):
                 self.pic = kwargs.get('pic', False)
                 if not isinstance(self.pic, bool):
                     raise InvalidArguments('Argument pic to static library {!r} must be boolean'.format(self.name))
+        self.implicit_include_directories = kwargs.get('implicit_include_directories', True)
+        if not isinstance(self.implicit_include_directories, bool):
+            raise InvalidArguments('Implicit_include_directories must be a boolean.')
 
     def get_filename(self):
         return self.filename
@@ -770,7 +831,7 @@ class BuildTarget(Target):
                 self.external_deps.append(extpart)
                 # Deps of deps.
                 self.add_deps(dep.ext_deps)
-            elif isinstance(dep, dependencies.Dependency):
+            elif isinstance(dep, dependencies.ExternalDependency):
                 self.external_deps.append(dep)
                 self.process_sourcelist(dep.get_sources())
             elif isinstance(dep, BuildTarget):
@@ -780,7 +841,7 @@ You probably should put it in link_with instead.''')
                 # This is a bit of a hack. We do not want Build to know anything
                 # about the interpreter so we can't import it and use isinstance.
                 # This should be reliable enough.
-                if hasattr(dep, 'args_frozen'):
+                if hasattr(dep, 'project_args_frozen') or hasattr(dep, 'global_args_frozen'):
                     raise InvalidArguments('Tried to use subproject object as a dependency.\n'
                                            'You probably wanted to use a dependency declared in it instead.\n'
                                            'Access it by calling get_variable() on the subproject object.')
@@ -793,13 +854,11 @@ You probably should put it in link_with instead.''')
         return self.external_deps
 
     def link(self, target):
-        if not isinstance(target, list):
-            target = [target]
-        for t in target:
+        for t in flatten(target):
             if hasattr(t, 'held_object'):
                 t = t.held_object
-            if not isinstance(t, (StaticLibrary, SharedLibrary)):
-                raise InvalidArguments('Link target {!r} is not library.'.format(t))
+            if not t.is_linkable_target():
+                raise InvalidArguments('Link target {!r} is not linkable.'.format(t))
             if isinstance(self, SharedLibrary) and isinstance(t, StaticLibrary) and not t.pic:
                 msg = "Can't link non-PIC static library {!r} into shared library {!r}. ".format(t.name, self.name)
                 msg += "Use the 'pic' option to static_library to build with PIC."
@@ -809,9 +868,7 @@ You probably should put it in link_with instead.''')
             self.link_targets.append(t)
 
     def link_whole(self, target):
-        if not isinstance(target, list):
-            target = [target]
-        for t in target:
+        for t in flatten(target):
             if hasattr(t, 'held_object'):
                 t = t.held_object
             if not isinstance(t, StaticLibrary):
@@ -825,7 +882,7 @@ You probably should put it in link_with instead.''')
             self.link_whole_targets.append(t)
 
     def add_pch(self, language, pchlist):
-        if len(pchlist) == 0:
+        if not pchlist:
             return
         elif len(pchlist) == 1:
             if not environment.is_header(pchlist[0]):
@@ -884,7 +941,7 @@ You probably should put it in link_with instead.''')
                 langs.append(dep.language)
         # Check if any of the internal libraries this target links to were
         # written in this language
-        for link_target in self.link_targets:
+        for link_target in itertools.chain(self.link_targets, self.link_whole_targets):
             for language in link_target.compilers:
                 if language not in langs:
                     langs.append(language)
@@ -911,7 +968,14 @@ You probably should put it in link_with instead.''')
         # Pick a compiler based on the language priority-order
         for l in clike_langs:
             if l in self.compilers or l in dep_langs:
-                return all_compilers[l]
+                try:
+                    return all_compilers[l]
+                except KeyError:
+                    raise MesonException(
+                        'Could not get a dynamic linker for build target {!r}. '
+                        'Requires a linker for language "{}", but that is not '
+                        'a project language.'.format(self.name, l))
+
         m = 'Could not get a dynamic linker for build target {!r}'
         raise AssertionError(m.format(self.name))
 
@@ -938,6 +1002,9 @@ You probably should put it in link_with instead.''')
             return True
         return False
 
+    def is_linkable_target(self):
+        return False
+
 
 class Generator:
     def __init__(self, args, kwargs):
@@ -1065,7 +1132,7 @@ class Executable(BuildTarget):
         if not hasattr(self, 'suffix'):
             # Executable for Windows or C#/Mono
             if (for_windows(is_cross, environment) or
-                for_cygwin(is_cross, environment) or 'cs' in self.compilers):
+                    for_cygwin(is_cross, environment) or 'cs' in self.compilers):
                 self.suffix = 'exe'
             else:
                 self.suffix = ''
@@ -1074,9 +1141,52 @@ class Executable(BuildTarget):
             self.filename += '.' + self.suffix
         self.outputs = [self.filename]
 
+        # The import library this target will generate
+        self.import_filename = None
+        # The import library that Visual Studio would generate (and accept)
+        self.vs_import_filename = None
+        # The import library that GCC would generate (and prefer)
+        self.gcc_import_filename = None
+
+        # if implib appears, this target is linkwith:-able, but that only means
+        # something on Windows platforms.
+        self.is_linkwithable = False
+        if 'implib' in kwargs and kwargs['implib']:
+            implib_basename = self.name + '.exe'
+            if not isinstance(kwargs['implib'], bool):
+                implib_basename = kwargs['implib']
+            self.is_linkwithable = True
+            if for_windows(is_cross, environment) or for_cygwin(is_cross, environment):
+                self.vs_import_filename = '{0}.lib'.format(implib_basename)
+                self.gcc_import_filename = 'lib{0}.a'.format(implib_basename)
+
+                if self.get_using_msvc():
+                    self.import_filename = self.vs_import_filename
+                else:
+                    self.import_filename = self.gcc_import_filename
+
     def type_suffix(self):
         return "@exe"
 
+    def check_unknown_kwargs(self, kwargs):
+        self.check_unknown_kwargs_int(kwargs, known_exe_kwargs)
+
+    def get_import_filename(self):
+        """
+        The name of the import library that will be outputted by the compiler
+
+        Returns None if there is no import library required for this platform
+        """
+        return self.import_filename
+
+    def get_import_filenameslist(self):
+        if self.import_filename:
+            return [self.vs_import_filename, self.gcc_import_filename]
+        return []
+
+    def is_linkable_target(self):
+        return self.is_linkwithable
+
 class StaticLibrary(BuildTarget):
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         if 'pic' not in kwargs and 'b_staticpic' in environment.coredata.base_options:
@@ -1084,6 +1194,14 @@ class StaticLibrary(BuildTarget):
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         if 'cs' in self.compilers:
             raise InvalidArguments('Static libraries not supported for C#.')
+        if 'rust' in self.compilers:
+            # If no crate type is specified, or it's the generic lib type, use rlib
+            if not hasattr(self, 'rust_crate_type') or self.rust_crate_type == 'lib':
+                mlog.debug('Defaulting Rust static library target crate type to rlib')
+                self.rust_crate_type = 'rlib'
+            # Don't let configuration proceed with a non-static crate type
+            elif self.rust_crate_type not in ['rlib', 'staticlib']:
+                raise InvalidArguments('Crate type "{0}" invalid for static libraries; must be "rlib" or "staticlib"'.format(self.rust_crate_type))
         # By default a static library is named libfoo.a even on Windows because
         # MSVC does not have a consistent convention for what static libraries
         # are called. The MSVC CRT uses libfoo.lib syntax but nothing else uses
@@ -1094,9 +1212,12 @@ class StaticLibrary(BuildTarget):
         if not hasattr(self, 'prefix'):
             self.prefix = 'lib'
         if not hasattr(self, 'suffix'):
-            # Rust static library crates have .rlib suffix
             if 'rust' in self.compilers:
-                self.suffix = 'rlib'
+                if not hasattr(self, 'rust_crate_type') or self.rust_crate_type == 'rlib':
+                    # default Rust static library suffix
+                    self.suffix = 'rlib'
+                elif self.rust_crate_type == 'staticlib':
+                    self.suffix = 'a'
             else:
                 self.suffix = 'a'
         self.filename = self.prefix + self.name + '.' + self.suffix
@@ -1108,6 +1229,18 @@ class StaticLibrary(BuildTarget):
     def check_unknown_kwargs(self, kwargs):
         self.check_unknown_kwargs_int(kwargs, known_lib_kwargs)
 
+    def process_kwargs(self, kwargs, environment):
+        super().process_kwargs(kwargs, environment)
+        if 'rust_crate_type' in kwargs:
+            rust_crate_type = kwargs['rust_crate_type']
+            if isinstance(rust_crate_type, str):
+                self.rust_crate_type = rust_crate_type
+            else:
+                raise InvalidArguments('Invalid rust_crate_type "{0}": must be a string.'.format(rust_crate_type))
+
+    def is_linkable_target(self):
+        return True
+
 class SharedLibrary(BuildTarget):
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.soversion = None
@@ -1120,6 +1253,14 @@ class SharedLibrary(BuildTarget):
         # The import library that GCC would generate (and prefer)
         self.gcc_import_filename = None
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
+        if 'rust' in self.compilers:
+            # If no crate type is specified, or it's the generic lib type, use dylib
+            if not hasattr(self, 'rust_crate_type') or self.rust_crate_type == 'lib':
+                mlog.debug('Defaulting Rust dynamic library target crate type to "dylib"')
+                self.rust_crate_type = 'dylib'
+            # Don't let configuration proceed with a non-dynamic crate type
+            elif self.rust_crate_type not in ['dylib', 'cdylib']:
+                raise InvalidArguments('Crate type "{0}" invalid for dynamic libraries; must be "dylib" or "cdylib"'.format(self.rust_crate_type))
         if not hasattr(self, 'prefix'):
             self.prefix = None
         if not hasattr(self, 'suffix'):
@@ -1161,12 +1302,6 @@ class SharedLibrary(BuildTarget):
             prefix = ''
             suffix = 'dll'
             self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
-        # Rust
-        elif 'rust' in self.compilers:
-            # Currently, we always build --crate-type=rlib
-            prefix = 'lib'
-            suffix = 'rlib'
-            self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
         # C, C++, Swift, Vala
         # Only Windows uses a separate import library for linking
         # For all other targets/platforms import_filename stays None
@@ -1255,22 +1390,33 @@ class SharedLibrary(BuildTarget):
         # Visual Studio module-definitions file
         if 'vs_module_defs' in kwargs:
             path = kwargs['vs_module_defs']
+            if hasattr(path, 'held_object'):
+                path = path.held_object
             if isinstance(path, str):
                 if os.path.isabs(path):
                     self.vs_module_defs = File.from_absolute_file(path)
                 else:
                     self.vs_module_defs = File.from_source_file(environment.source_dir, self.subdir, path)
-                # link_depends can be an absolute path or relative to self.subdir
-                self.link_depends.append(path)
+                self.link_depends.append(self.vs_module_defs)
             elif isinstance(path, File):
                 # When passing a generated file.
                 self.vs_module_defs = path
-                # link_depends can be an absolute path or relative to self.subdir
-                self.link_depends.append(path.absolute_path(environment.source_dir, environment.build_dir))
+                self.link_depends.append(path)
+            elif hasattr(path, 'get_filename'):
+                # When passing output of a Custom Target
+                path = File.from_built_file(path.subdir, path.get_filename())
+                self.vs_module_defs = path
+                self.link_depends.append(path)
             else:
                 raise InvalidArguments(
                     'Shared library vs_module_defs must be either a string, '
-                    'or a file object')
+                    'a file object or a Custom Target')
+        if 'rust_crate_type' in kwargs:
+            rust_crate_type = kwargs['rust_crate_type']
+            if isinstance(rust_crate_type, str):
+                self.rust_crate_type = rust_crate_type
+            else:
+                raise InvalidArguments('Invalid rust_crate_type "{0}": must be a string.'.format(rust_crate_type))
 
     def check_unknown_kwargs(self, kwargs):
         self.check_unknown_kwargs_int(kwargs, known_lib_kwargs)
@@ -1324,6 +1470,9 @@ class SharedLibrary(BuildTarget):
     def type_suffix(self):
         return "@sha"
 
+    def is_linkable_target(self):
+        return True
+
 # A shared library that is meant to be used with dlopen rather than linking
 # into something else.
 class SharedModule(SharedLibrary):
@@ -1417,9 +1566,12 @@ class CustomTarget(Target):
 
     def process_kwargs(self, kwargs):
         super().process_kwargs(kwargs)
-        self.sources = kwargs.get('input', [])
-        if not isinstance(self.sources, list):
-            self.sources = [self.sources]
+        sources = flatten(kwargs.get('input', []))
+        self.sources = []
+        for s in sources:
+            if hasattr(s, 'held_object'):
+                s = s.held_object
+            self.sources.append(s)
         if 'output' not in kwargs:
             raise InvalidArguments('Missing keyword argument "output".')
         self.outputs = kwargs['output']
