@@ -21,11 +21,10 @@ from io import StringIO
 from ast import literal_eval
 from enum import Enum
 import tempfile
-import mesontest
+from mesonbuild import mtest
 from mesonbuild import environment
 from mesonbuild import mesonlib
 from mesonbuild import mlog
-from mesonbuild import mesonmain
 from mesonbuild.mesonlib import stringlistify, Popen_safe
 from mesonbuild.coredata import backendlist
 import argparse
@@ -184,7 +183,7 @@ def get_relative_files_list_from_dir(fromdir):
             paths.append(path)
     return paths
 
-def platform_fix_name(fname):
+def platform_fix_name(fname, compiler):
     if '?lib' in fname:
         if mesonlib.is_cygwin():
             fname = re.sub(r'\?lib(.*)\.dll$', r'cyg\1.dll', fname)
@@ -196,6 +195,16 @@ def platform_fix_name(fname):
         if mesonlib.is_windows() or mesonlib.is_cygwin():
             return fname + '.exe'
 
+    if fname.startswith('?msvc:'):
+        fname = fname[6:]
+        if compiler != 'cl':
+            return None
+
+    if fname.startswith('?gcc:'):
+        fname = fname[5:]
+        if compiler == 'cl':
+            return None
+
     return fname
 
 def validate_install(srcdir, installdir, compiler):
@@ -211,10 +220,13 @@ def validate_install(srcdir, installdir, compiler):
     elif os.path.exists(info_file):
         with open(info_file) as f:
             for line in f:
-                expected[platform_fix_name(line.strip())] = False
+                line = platform_fix_name(line.strip(), compiler)
+                if line:
+                    expected[line] = False
     # Check if expected files were found
     for fname in expected:
-        if os.path.exists(os.path.join(installdir, fname)):
+        file_path = os.path.join(installdir, fname)
+        if os.path.exists(file_path) or os.path.islink(file_path):
             expected[fname] = True
     for (fname, found) in expected.items():
         if not found:
@@ -249,6 +261,23 @@ def log_text_file(logfile, testdir, stdo, stde):
         executor.shutdown()
         raise StopException()
 
+
+def bold(text):
+    return mlog.bold(text).get_text(mlog.colorize_console)
+
+
+def green(text):
+    return mlog.green(text).get_text(mlog.colorize_console)
+
+
+def red(text):
+    return mlog.red(text).get_text(mlog.colorize_console)
+
+
+def yellow(text):
+    return mlog.yellow(text).get_text(mlog.colorize_console)
+
+
 def run_test_inprocess(testdir):
     old_stdout = sys.stdout
     sys.stdout = mystdout = StringIO()
@@ -258,12 +287,12 @@ def run_test_inprocess(testdir):
     os.chdir(testdir)
     test_log_fname = 'meson-logs/testlog.txt'
     try:
-        returncode_test = mesontest.run(['--no-rebuild'])
+        returncode_test = mtest.run(['--no-rebuild'])
         if os.path.exists(test_log_fname):
             test_log = open(test_log_fname, errors='ignore').read()
         else:
             test_log = ''
-        returncode_benchmark = mesontest.run(['--no-rebuild', '--benchmark', '--logbase', 'benchmarklog'])
+        returncode_benchmark = mtest.run(['--no-rebuild', '--benchmark', '--logbase', 'benchmarklog'])
     finally:
         sys.stdout = old_stdout
         sys.stderr = old_stderr
@@ -391,14 +420,17 @@ def have_d_compiler():
 def have_objc_compiler():
     with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
         env = environment.Environment(None, build_dir, None, get_fake_options('/'), [])
-        objc_comp = env.detect_objc_compiler(False)
+        try:
+            objc_comp = env.detect_objc_compiler(False)
+        except:
+            return False
         if not objc_comp:
             return False
         try:
             objc_comp.sanity_check(env.get_scratch_dir(), env)
+            objcpp_comp = env.detect_objc_compiler(False)
         except:
             return False
-        objcpp_comp = env.detect_objc_compiler(False)
         if not objcpp_comp:
             return False
         try:
@@ -446,10 +478,15 @@ def detect_tests_to_run():
     return gathered_tests
 
 def run_tests(all_tests, log_name_base, extra_args):
-    global stop, executor, futures
+    global logfile
     txtname = log_name_base + '.txt'
+    with open(txtname, 'w', encoding="utf_8") as lf:
+        logfile = lf
+        return _run_tests(all_tests, log_name_base, extra_args)
+
+def _run_tests(all_tests, log_name_base, extra_args):
+    global stop, executor, futures
     xmlname = log_name_base + '.xml'
-    logfile = open(txtname, 'w', encoding="utf_8")
     junit_root = ET.Element('testsuites')
     conf_time = 0
     build_time = 0
@@ -474,10 +511,12 @@ def run_tests(all_tests, log_name_base, extra_args):
 
     for name, test_cases, skipped in all_tests:
         current_suite = ET.SubElement(junit_root, 'testsuite', {'name': name, 'tests': str(len(test_cases))})
+        print()
         if skipped:
-            print('\nNot running %s tests.\n' % name)
+            print(bold('Not running %s tests.' % name))
         else:
-            print('\nRunning %s tests.\n' % name)
+            print(bold('Running %s tests.' % name))
+        print()
         futures = []
         for t in test_cases:
             # Jenkins screws us over by automatically sorting test cases by name
@@ -493,7 +532,7 @@ def run_tests(all_tests, log_name_base, extra_args):
             sys.stdout.flush()
             result = result.result()
             if result is None or 'MESON_SKIP_TEST' in result.stdo:
-                print('Skipping:', t)
+                print(yellow('Skipping:'), t)
                 current_test = ET.SubElement(current_suite, 'testcase', {'name': testname,
                                                                          'classname': name})
                 ET.SubElement(current_test, 'skipped', {})
@@ -501,7 +540,7 @@ def run_tests(all_tests, log_name_base, extra_args):
             else:
                 without_install = "" if len(install_commands) > 0 else " (without install)"
                 if result.msg != '':
-                    print('Failed test{} during {}: {!r}'.format(without_install, result.step.name, t))
+                    print(red('Failed test{} during {}: {!r}'.format(without_install, result.step.name, t)))
                     print('Reason:', result.msg)
                     failing_tests += 1
                     if result.step == BuildStep.configure and result.mlog != no_meson_log_msg:
@@ -647,9 +686,9 @@ if __name__ == '__main__':
         pass
     for f in pbfiles:
         os.unlink(f)
-    print('\nTotal passed tests:', passing_tests)
-    print('Total failed tests:', failing_tests)
-    print('Total skipped tests:', skipped_tests)
+    print('\nTotal passed tests:', green(str(passing_tests)))
+    print('Total failed tests:', red(str(failing_tests)))
+    print('Total skipped tests:', yellow(str(skipped_tests)))
     if failing_tests > 0:
         print('\nMesonlogs of failing tests\n')
         for l in failing_logs:
