@@ -23,6 +23,7 @@ import tempfile
 import platform
 from mesonbuild import mesonlib
 from mesonbuild import mesonmain
+from mesonbuild import mlog
 from mesonbuild.environment import detect_ninja
 from io import StringIO
 from enum import Enum
@@ -118,7 +119,7 @@ def get_fake_options(prefix):
     return opts
 
 def should_run_linux_cross_tests():
-    return shutil.which('arm-linux-gnueabihf-gcc-6') and not platform.machine().startswith('arm')
+    return shutil.which('arm-linux-gnueabihf-gcc-6') and not platform.machine().lower().startswith('arm')
 
 def run_configure_inprocess(commandlist):
     old_stdout = sys.stdout
@@ -126,7 +127,7 @@ def run_configure_inprocess(commandlist):
     old_stderr = sys.stderr
     sys.stderr = mystderr = StringIO()
     try:
-        returncode = mesonmain.run(commandlist[0], commandlist[1:])
+        returncode = mesonmain.run(commandlist[1:], commandlist[0])
     finally:
         sys.stdout = old_stdout
         sys.stderr = old_stderr
@@ -141,7 +142,17 @@ class FakeEnvironment(object):
     def is_cross_build(self):
         return False
 
+def print_system_info():
+    print(mlog.bold('System information.').get_text(mlog.colorize_console))
+    print('Architecture:', platform.architecture())
+    print('Machine:', platform.machine())
+    print('Platform:', platform.system())
+    print('Processor:', platform.processor())
+    print('System:', platform.system())
+    print('')
+
 if __name__ == '__main__':
+    print_system_info()
     # Enable coverage early...
     enable_coverage = '--cov' in sys.argv
     if enable_coverage:
@@ -177,14 +188,8 @@ if __name__ == '__main__':
     if 'APPVEYOR' in os.environ and os.environ['arch'] == 'x86':
         os.environ.pop('platform')
     # Run tests
-    print('Running unittests.\n')
-    units = ['InternalTests', 'AllPlatformTests', 'FailureTests']
-    if mesonlib.is_linux():
-        units += ['LinuxlikeTests']
-        if should_run_linux_cross_tests():
-            units += ['LinuxArmCrossCompileTests']
-    elif mesonlib.is_windows():
-        units += ['WindowsTests']
+    print(mlog.bold('Running unittests.').get_text(mlog.colorize_console))
+    print()
     # Can't pass arguments to unit tests, so set the backend to use in the environment
     env = os.environ.copy()
     env['MESON_UNIT_TEST_BACKEND'] = backend.name
@@ -196,11 +201,11 @@ if __name__ == '__main__':
                         'coverage.process_startup()\n')
             env['COVERAGE_PROCESS_START'] = '.coveragerc'
             env['PYTHONPATH'] = os.pathsep.join([td] + env.get('PYTHONPATH', []))
-
-        returncode += subprocess.call([sys.executable, 'run_unittests.py', '-v'] + units, env=env)
+        returncode += subprocess.call([sys.executable, 'run_unittests.py', '-v'], env=env)
         # Ubuntu packages do not have a binary without -6 suffix.
         if should_run_linux_cross_tests():
-            print('Running cross compilation tests.\n')
+            print(mlog.bold('Running cross compilation tests.').get_text(mlog.colorize_console))
+            print()
             returncode += subprocess.call([sys.executable, 'run_cross_test.py', 'cross/ubuntu-armhf.txt'], env=env)
         returncode += subprocess.call([sys.executable, 'run_project_tests.py'] + sys.argv[1:], env=env)
     sys.exit(returncode)
