@@ -813,6 +813,8 @@ class AllPlatformTests(BasePlatformTests):
         self._run(self.mtest_command + ['--setup=empty'])
         # Setup with only env works
         self._run(self.mtest_command + ['--setup=onlyenv'])
+        self._run(self.mtest_command + ['--setup=onlyenv2'])
+        self._run(self.mtest_command + ['--setup=onlyenv3'])
         # Setup with only a timeout works
         self._run(self.mtest_command + ['--setup=timeout'])
 
@@ -897,7 +899,7 @@ class AllPlatformTests(BasePlatformTests):
             raise Exception('Could not find someexe and somfxe commands')
         # Check include order for 'someexe'
         incs = [a for a in shlex.split(execmd) if a.startswith("-I")]
-        self.assertEqual(len(incs), 8)
+        self.assertEqual(len(incs), 9)
         # target private dir
         self.assertPathEqual(incs[0], "-Isub4/someexe@exe")
         # target build subdir
@@ -914,6 +916,8 @@ class AllPlatformTests(BasePlatformTests):
         self.assertPathEqual(incs[6], "-Isub1")
         # target internal dependency include_directories: source dir
         self.assertPathBasenameEqual(incs[7], 'sub1')
+        # custom target include dir
+        self.assertPathEqual(incs[8], '-Ictsub')
         # Check include order for 'somefxe'
         incs = [a for a in shlex.split(fxecmd) if a.startswith('-I')]
         self.assertEqual(len(incs), 9)
@@ -1288,6 +1292,9 @@ int main(int argc, char **argv) {
         targets = mesonbuild.coredata.forbidden_target_names
         # We don't actually define a target with this name
         targets.pop('build.ninja')
+        # Remove this to avoid multiple entries with the same name
+        # but different case.
+        targets.pop('PHONY')
         for i in targets:
             self.assertPathExists(os.path.join(testdir, i))
 
@@ -1907,6 +1914,19 @@ class LinuxlikeTests(BasePlatformTests):
         for i in compdb:
             self.assertIn("-fsanitize=address", i["command"])
 
+    def test_coverage(self):
+        if not shutil.which('gcovr'):
+            raise unittest.SkipTest('gcovr not found')
+        if not shutil.which('genhtml'):
+            raise unittest.SkipTest('genhtml not found')
+        if 'clang' in os.environ.get('CC', '') and os.environ.get('TRAVIS_OS_NAME', '') == 'linux':
+            raise unittest.SkipTest('Gcovr has a bug and does not work with Clang in the CI environment.')
+        testdir = os.path.join(self.common_test_dir, '1 trivial')
+        self.init(testdir, ['-Db_coverage=true'])
+        self.build()
+        self.run_tests()
+        self.run_target('coverage-html')
+
 class LinuxArmCrossCompileTests(BasePlatformTests):
     '''
     Tests that verify cross-compilation to Linux/ARM
